/*
killall -9 smlprt
make
./smlprt
echo "-MAKE SCRIPT DONE-" 
sleep 1
*/

/*****************************************************
SMLPRT
Small Particle Gravity Simulator
by Chris Merck <navaburo@gmail.com>
Compiles best on Linux (but may compile on other systems)

   QBasic Build: March 2003
   C Build: May 2003
   Linux C build(s): April 2006

This software may be freely redistributed under the terms
 of the GNU public license.

 png code by Guillaume Cottenceau
 
 base SDL code taken from:
 Andrew M. <andrew@textux.com>
*****************************************************/ 


#include <unistd.h>
#include <string.h>
#include <stdarg.h>
//#define PNG_DEBUG 3
//#include <png.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <SDL.h>




#define FONT_HEIGHT  8
#define FONT_WIDTH   8

//define edges of screen
#define XMAX 640
#define YMAX 480

#define  NUMPRT 1000

#define  TAILLEN 15


#define FPS_DISP_TIMER_MAX 100

#define DISTMODMIN .6


#define FDIV 10000

//SFACTOR is the global energy retention per frame
#define SFACTOR .99
//RUBFACTOR is the energy retention when within distmod of another particle
#define RUBFACTOR .9999
#define RUBDIST 50

#define WALLRUBFACTOR .99

//this pulls the particles down uniformly
	#define GRAVITY .01

#define INTERATTRACT_CUTOFF_FACTOR 2

#define DEFAULT_FUNCTION 3

void setpixel(SDL_Surface *screen, int x, int y, Uint8 r, Uint8 g, Uint8 b)
{
  Uint8 *ubuff8;
  Uint16 *ubuff16;
  Uint32 *ubuff32;
  Uint32 color;
  char c1, c2, c3;
  
  /* Lock the screen, if needed */
  if(SDL_MUSTLOCK(screen)) {
    if(SDL_LockSurface(screen) < 0) 
      return;
  }
  
  /* Get the color */
  color = SDL_MapRGB( screen->format, r, g, b );
  
  /* How we draw the pixel depends on the bitdepth */
  switch(screen->format->BytesPerPixel) 
    {
    case 1: 
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + x; 
      *ubuff8 = (Uint8) color;
      break;

    case 2:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + (x*2);
      ubuff16 = (Uint16*) ubuff8;
      *ubuff16 = (Uint16) color; 
      break;  

    case 3:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + (x*3);
      

      if(SDL_BYTEORDER == SDL_LIL_ENDIAN) {
	c1 = (color & 0xFF0000) >> 16;
	c2 = (color & 0x00FF00) >> 8;
	c3 = (color & 0x0000FF);
      } else {
	c3 = (color & 0xFF0000) >> 16;
	c2 = (color & 0x00FF00) >> 8;
	c1 = (color & 0x0000FF);	
      }

      ubuff8[0] = c3;
      ubuff8[1] = c2;
      ubuff8[2] = c1;
      break;
      
    case 4:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y*screen->pitch) + (x*4);
      ubuff32 = (Uint32*)ubuff8;
      *ubuff32 = color;
      break;
      
    default:
      fprintf(stderr, "Error: Unknown bitdepth!\n");
    }
  
  /* Unlock the screen if needed */
  if(SDL_MUSTLOCK(screen)) {
    SDL_UnlockSurface(screen);
  }
}

void iniprt(double *x, double *y)
{
  double tmp1, tmp2;

   /* Erase old pixel */
   //vga_setcolor(0);
   //vga_drawpixel(*x,*y);

   /* Initialize/reset the particle position */
   
   //this tmp crap is unnessisary. but HOW DO I CONVERT int to double otherwise???
   tmp1=rand();
   tmp2=RAND_MAX;
   tmp1=tmp1/tmp2;
   tmp1=200*tmp1;
   *x=tmp1;
   
   tmp1=rand();
   tmp2=RAND_MAX;
   tmp1=tmp1/tmp2;
   tmp1=200*tmp1;
   *y=tmp1;
}


int main(int argc, char* argv[])
{
  /* Declare Variables */
  SDL_Surface *screen;
  SDL_Event event;
  
  int exitkey = 0;
  
  	struct timeval {
		time_t         tv_sec;        /* seconds */
		suseconds_t    tv_usec;  /* microseconds */
	};
	
	int taxidist;
	int mx, my, rx;
	int r,g,b;
	double tp;
	int walls=1; //this controls the existance of walls
	double PWR=10;
	int dffunc=DEFAULT_FUNCTION;
	double SRAMP=.5;
	double SCALE=15;
	double SPDMAX=10;
	int lastkey;
	double tx, ty, txs, tys;
 	int np,n,m,tox, toy;
	double distmod;
	double tmp;
   	double dist, f;
	double x[NUMPRT], y[NUMPRT], xs[NUMPRT], ys[NUMPRT];
	double oxs[NUMPRT], oys[NUMPRT], oxa[NUMPRT], oya[NUMPRT];
	int ix, iy;
	int friends[NUMPRT];
	double dimm;
	struct timeval tim;
	double endtime=0, oendtime=0, fps, ofps;
	int fpsdisptimer=FPS_DISP_TIMER_MAX;
	char *fpsstring;
	FILE *fpdebug;
	
	printf("Small Partical Gravity Simulator\nBy Chris Merck\n");
	printf("This build requires SDL.\n Contact navaburo@gmail.com for more info\n");
	
	/*do {
	printf("\nNumber of Particals(max=%d):",NUMPRT);
   	scanf("%d",&np);
        } while (np>NUMPRT);*/
	np=300; //TODO: make this a command-line argument

	//initialize variables from command-line arguments
	/*if (argc<6)
	{
		printf("usage: pix silenti visablei cmin cmax amin amax\n");
		return(1);
	}
	SILENTI = atoi(argv[1]);
	VISABLEI = atoi(argv[2]);
	cmin = atof(argv[3]);
	cmax = atof(argv[4]);
	amin = atof(argv[5]);
	amax = atof(argv[6]);
	*/	
	
	//clear debug file
	system("rm debug.out");
	
	//open debug dump file
	if ((fpdebug = fopen("debug.out", "a+")) == NULL)
 	{
		printf("Cannot open debug file. \n");
		return(1);
	}
	
	srand(1);  /* Fetch a set of fresh random numbers */ //TODO: get a time based seed
	
  /* Initialize SDL, exit if there is an error. */
  if( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Could not initialize SDL: %s\n", 
	    SDL_GetError());
    return -1;
  }
  
  /* When the program is through executing, call SDL_Quit */
  atexit(SDL_Quit);
  
  /* Grab a surface on the screen */
  screen = SDL_SetVideoMode(640, 480, 32, SDL_SWSURFACE|SDL_ANYFORMAT);
  if( !screen ) {
    fprintf(stderr, "Couldn't create a surface: %s\n",
	    SDL_GetError());
    return -1;
  }
  
  /* Print Some info */
  printf("Created SDL Surface: %x\n", screen);
  printf("\t BPP \t\t : %d\n", screen->format->BytesPerPixel );
  printf("\t XRes \t\t :%d\n", screen->w );
  printf("\t YRes \t\t :%d\n", screen->h );
  
  /*
  while(!exitkey) {
    DrawPicture(screen);
    while(SDL_PollEvent(&event)) {
      switch( event.type ) 
	{
	case SDL_QUIT:
	  exitkey = 1;
	  printf("Got quit event!\n");
	  break;
	case SDL_KEYDOWN:
	  printf("Key hit - exiting!\n");
	  exitkey = 1;
	  break;
	}
    }
  }
  */
  	
	//display welcome message on screen
	/*tx=XMAX/2; ty=100;
	gl_write(tx, ty, "Small Particle Gravitation Simulator"); ty+=10;
	gl_write(tx, ty, "by Chris Merck"); ty+=10;
	gl_write(tx, ty, "navaburo@gmail.com"); ty+=20;
	gl_write(tx, ty, "Use W and S to change the distance scale");
	*/
	
	printf("initializing particles...\n");
	
	for (n=1; n!=np; n++) 
	{
	iniprt(&x[n],&y[n]); /* Initialize Particle Positions */
	//fprintf(fpdebug, "\ninix=%1.2f\n", x[n]);
	}
	
	printf("begining main loop...\n");
	
   	/* START OF MAIN PROGRAM LOOP */
	do {
		for (n=1;n!=np;n++) {
			for (m=1;m!=np;m++) {
				if (n==m) continue; /* Don't pull itself! */
		
				/* Calculate distance */
				dist = sqrt(pow((x[m] - x[n]),2) + pow((y[m]-y[n]),2));

				if (dist<INTERATTRACT_CUTOFF_FACTOR*SCALE)
				{
				
				
				/* Calculate force exerted */
				distmod=dist/SCALE; //distbig 
				if (distmod<DISTMODMIN) distmod=DISTMODMIN;
				switch (dffunc)
				{
				case 0:
					//regular gravitation
					f = 50 / pow(dist,3);
					break;
				case 1:
					//repel when close, attract when far
					f = ((2*distmod-2)/distmod+2) / pow(distmod,PWR)/FDIV;
					break;
				case 2:
					//sinesoidal field (strange)
					f = -cos(distmod)/FDIV/1000.;
					break;
				case 3:
					//double hump (crude)
					f = -cos(distmod)/pow(distmod,PWR)/FDIV;
					break;
				case 4:
					//revised double hump
					f=-(PWR/10./(pow(distmod,2)+1.)-1./(5.*pow(distmod,2) + .3)-1./(5.*(pow(distmod-1.,2)+.5)))/FDIV;
				}
				
				if (dist<100) friends[n]++;
				//if (dist<20) friends[n]++;
				
				if (dist<SCALE*SRAMP) {
				//rub-off speed
				xs[n]=(99.*xs[n] + 1.*xs[m])/100.;
				ys[n]=(99.*ys[n] + 1.*ys[m])/100.;
				
				//rub off heat
				xs[n]*=RUBFACTOR;
				ys[n]*=RUBFACTOR;
				}
						
				/* Influence object in correct direction */
				xs[n] += f*(x[m]-x[n]);
				ys[n] += f*(y[m]-y[n]);
				}
				
			} /* next M */
	
			//exert uniform (planitary) gravitation downwards
			ys[n] += GRAVITY;
			
			//repel from mouse
			dist = sqrt(pow((mx - x[n]),2) + pow((my-y[n]),2));
			if (dist<80) {
				f = 2. / pow(dist,2);
				xs[n] += f*(mx-x[n]);
				ys[n] += f*(my-y[n]);
			}
			
			/* Check for off screen, and bounce if walls, respawn if no walls */
			if (y[n] < 0)
			{
				//if (walls == -1){
				//	iniprt(&x[n],&y[n]);
				//	xs[n]=0; ys[n]=0;
				//} else {
				//	ys[n]=abs(ys[n]);
				//	y[n]=.5;
				//}
				
			}
			if (x[n] > XMAX)
			{
				if (walls == -1){
					iniprt(&x[n],&y[n]);
					xs[n]=0; ys[n]=0;
				} else {
					xs[n]=-abs(xs[n]);
					x[n]=XMAX-.5;
				}
			}	
			if ( x[n] < 0 )
			{
				if (walls == -1){
					iniprt(&x[n],&y[n]);
					xs[n]=0; ys[n]=0;
				} else {
					xs[n]=abs(xs[n]);
					x[n]=.5;
				}
			}
			if (y[n] > YMAX)
			{
				y[n] = YMAX - 1; //place back on screen
				ys[n] = 0; //slow down
			}	
			
			//the floor gives a repulsive force
			dist = abs(YMAX - y[n] - 250);
			if (dist < 50 && x[n] < XMAX/2) f = -.1;
			
			/*
			distmod=dist/SCALE;
			if (distmod<DISTMODMIN) distmod=DISTMODMIN;
			f = 20./pow(distmod,2);
			*/
			ys[n] += f;
		        
	
			/* Save old values, move the part and DRAW THAT PIXEL!!! */
			/*ox[n][TAILLEN] = x[n];
			oy[n][TAILLEN]= y[n];*/
	
			/* Save tail data */
			/*for (m=TAILLEN;m>=0;m--) {
				ox[n][TAILLEN-m]=ox[n][TAILLEN-m-1];
				oy[n][TAILLEN-m]=oy[n][TAILLEN-m-1];
			}*/

			//cap particle speed
			/*if (xs[n]>SPDMAX) xs[n]=SPDMAX;
			if (xs[n]<-SPDMAX) xs[n]=-SPDMAX;
			if (ys[n]>SPDMAX) ys[n]=SPDMAX;
			if (ys[n]<-SPDMAX) ys[n]=-SPDMAX;*/

			//lose some speed...
			xs[n]*=SFACTOR;
			ys[n]*=SFACTOR;
			
			//erase old particle
			//if (x[n] > 0 && x[n] < screen->w && y[n] > 0 && y[n] < screen->h)
			//	setpixel(screen,x[n],y[n],255,255,255);
			
			//move particle
			x[n] += xs[n];
			y[n] += ys[n];
			
			r = friends[n]*20;
			if (r>255) r=255;
			g=255-r;
			b=(ys[n]+xs[n])*2;
			if (b>255) b=255;
			
			
			
			
			//draw a tail
			tx = x[n];
			ty = y[n];
			txs = xs[n];
			tys = ys[n];
			for (tp = 1; tp <= 100; tp++) {
				txs += .5*(xs[n] - oxs[n] - oxa[n])*tp;
				tys += .5*(ys[n] - oys[n] - oya[n])*tp;
				tx -= txs;
				ty -= tys;
				
				taxidist = abs(tx-x[n])+abs(ty-y[n]);
				if (taxidist > 1000) goto notp;
				
				dimm = pow(tp, .5);
				
				if (tx > 0 && tx < screen->w && ty > 0 && ty < screen->h)
					setpixel(screen,tx,ty,r/dimm,g/dimm,b/dimm);
			}
			
			notp:
			
			//draw new particle
			if (x[n] > 0 && x[n] < screen->w && y[n] > 0 && y[n] < screen->h)
				setpixel(screen,x[n],y[n],r,g,b);
			

			oxa[n] = (oxa[n] + xs[n] - oxs[n])/2.; //record old acceleration
			oya[n] = (oya[n] + ys[n] - oys[n])/2.;
			oxs[n] = xs[n]; //record old speeds
			oys[n] = ys[n];
			
			//printf ("%1.2f\n",oxa[n]);
			
			friends[n]=0;
		}
		
	//make the changes to the screen visable
	SDL_Flip(screen); 
	
	//clear screen
	SDL_FillRect( screen, NULL, SDL_MapRGB(screen->format,0,0,0) );
	
	
	
	
	//determine frame rate
	oendtime=endtime;
	gettimeofday(&tim, NULL);
	endtime=tim.tv_sec+(tim.tv_usec/1000000.0); //make endtime contain the time in seconds
	fps=(fps + 1./(endtime-oendtime))/2.; //smooth out by averaging last and current fps
	if (--fpsdisptimer < 0) //only display the fps every so often so we can read it
	{
		//sprintf(fpsstring, "%1.0f", fps);
		//printf("determining fps1\n");
		fpsdisptimer = FPS_DISP_TIMER_MAX;
		//printf("determining fps2\n");
		printf("FPS:%1.0f\n", fps);
	}
	
	

    	SDL_PumpEvents();
    	SDL_GetMouseState(&mx, &my);
	
	/*
	//erase old mouse dot
	vga_setcolor(0);
	vga_drawpixel(mx,my);
	//get the mouse position
	mouse_update();
        mouse_getposition_6d(&mx, &my, NULL, &rx, NULL, NULL);
	//draw a dot as a mouse cursor
	vga_setcolor(3);
	vga_drawpixel(mx,my);
	*/
	
	
	/*
	//lastkey=vga_getkey();
	if (lastkey==119) SCALE++; //w
	if (lastkey==115) SCALE--; //s
	if (lastkey==97)  PWR--; //a
	if (lastkey==100) PWR++; //d
	//if (lastkey==99) vga_clear(); //c = refresh
	if (lastkey==98) walls=-walls; //b = toggle walls on/off
	*/
	
	
   	} while (lastkey!=32); //wait for spacebar
	
	printf("Got SPACEBAR, exiting...");
  
  return 0;
}
